(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2020 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`Hamiltonian`",
    {"ProcessOptions`", "AtomicDensityMatrix`", "AtomicDensityMatrix`WignerEckart`"}
];

Begin["`Private`"];

Options@InternalHamiltonian =
    Options@E1Hamiltonian =
        Options@MagneticHamiltonian =
            Options@PolarizabilityHamiltonian =
                FilterRules[Options[WignerEckart], Except@Representation];

InternalHamiltonian[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    WignerEckart[sys, {Energy, 0, 0}, opts];

Energies[sys_?AtomicSystemQ, crit_ : True] :=
    Union[WignerEckart[#, {Energy, 0, 0}, #]& /@ SelectStates[sys, crit]];

TransitionFrequencies[sys_?AtomicSystemQ, label1_ : True, label2_ : True] :=
    Transpose@Expand@Outer[Plus, Energies[sys, label2], -Energies[sys, label1]];

E1Hamiltonian[sys_?AtomicSystemQ /; MatchQ[AtomicSystemType@sys, "Toy" | "Hyperfine"],
    efield_ /; CartesianQ@efield || ScalarQ@efield, opts : OptionsPattern[]] :=
    Module[{components = DeleteCases[EnsureList@efield, _?PossibleZeroQ], field},
        field = Switch[
            Length@components,
            0, 0,
            _, First@components
        ];
        -ComplexExpand[Re@field WignerEckart[sys, {Dipole, 1, 0}, opts], TargetFunctions -> {Re, Im}]
    ];

E1Hamiltonian[sys_?AtomicSystemQ, efield_?CartesianQ, opts : OptionsPattern[]] :=
    -ComplexExpand[CenterDot[Re@efield, ToCartesian@WignerEckart[sys, {Dipole, 1}, opts]], TargetFunctions -> {Re, Im}];

MagneticHamiltonian[sys_?AtomicSystemQ /; MatchQ[AtomicSystemType@sys, "Toy" | "Hyperfine"], bf_, opts : OptionsPattern[]] :=
    WignerEckart[sys, {Zero, 0, 0}];

MagneticHamiltonian[sys_?AtomicSystemQ, bf_?CartesianQ, opts : OptionsPattern[]] :=
    CenterDot[-bf, ToCartesian@WignerEckart[sys, {MagneticMoment, 1}, opts]];

(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[*)(*Subscript[H, Stark]={*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"]\[CircleTimes]*)(*SuperscriptBox[*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"],"*"]}, \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

PolarizabilityHamiltonian[sys_?AtomicSystemQ, efield_?CartesianQ, k_Integer, opts : OptionsPattern[]] :=
    CenterDot[
        SphericalTensorProduct[ToCovariant@efield, ToCovariant@EC@efield, {k}],
        ECT@WignerEckart[sys, {Polarizability, k}, opts]
    ];

PolarizabilityHamiltonian[sys_?AtomicSystemQ, efield_?ScalarQ, k_Integer, opts : OptionsPattern[]] :=
    PolarizabilityHamiltonian[sys, {0, 0, efield}, k, opts];

PolarizabilityHamiltonian[sys_?AtomicSystemQ, efield_, opts : OptionsPattern[]] :=
    Sum[PolarizabilityHamiltonian[sys, efield, k, opts], {k, 0, 2}];

(*TransitionPolarizabilityHamiltonian[sys_?AtomicSystemQ, opticalfield_?CartesianQ, staticfield_?CartesianQ, k_] :=*)
    (*SphericalTensorProduct[*)
        (*ToCovariant@staticfield,*)
        (*ToCovariant@ComplexExpand@Re@opticalfield, {k}*)
    (*]\[CenterDot]ECT@WignerEckart[sys, {TransitionPolarizability, k}];*)

(*TransitionPolarizabilityHamiltonian[sys_?AtomicSystemQ, opticalfield_, staticfield_] :=*)
    (*Sum[TransitionPolarizabilityHamiltonian[sys, opticalfield, staticfield, k], {k, 0, 2}];*)

Options[Hamiltonian] = Join[{ElectricField -> {0, 0, 0}, MagneticField -> {0, 0, 0}, Interaction -> Automatic}, FilterRules[Options[WignerEckart], Except@Representation]];

RestrictCoupling = AllowedCouplings;

OptionProcessor[Interaction] =
    Switch[
        #,
        All,
        {"Internal", "ElectricDipole", "MagneticDipole", "Polarizability"},
        Automatic,
        {"Internal", "ElectricDipole", "MagneticDipole"},
        None,
        {},
        _,
      Flatten@{#}
    ] /. {
        Internal -> "Internal",
        E1 -> "ElectricDipole",
        Magnetic -> "MagneticDipole",
        Polarizability -> "Polarizability"
    }&;
OptionValidator[Interaction] = {("Internal" | "ElectricDipole" | "MagneticDipole" | "Polarizability") ...};

Hamiltonian[sys_?AtomicSystemQ, expSystem : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    Hamiltonian[sys, #, opts]& /@ expSystem;

OptionProcessor[OpticalParameters] :=
    Replace[# /.
        {
            _OpticalParameters -> {{0, 0}},
            x : {{_, Except[_Rule].., genericOptionsPattern[]}..} -> x,
            x : {_, Except[_Rule].., genericOptionsPattern[]} -> {x},
            x_ -> {{x}}
        },
        {params__, opts : genericOptionsPattern[]} :>
            {{params, FilterRules[{opts}, Options[OpticalField]]}, FilterRules[{opts}, Except@Options[OpticalField]]},
        {1}
    ]&;

Hamiltonian[sys_?AtomicSystemQ, region_ExperimentalRegion, opts : OptionsPattern[]] :=
    Hamiltonian[sys,
        ElectricField -> Join[
            If[DefinedQ[region, ElectricField], OptionProcessor[ElectricField][ElectricField@region], {{{0, 0, 0}}}],
            {OpticalField@@First@#, Rest@#}& /@ OptionProcessor[OpticalParameters][OpticalParameters@region]
        ],
        MagneticField -> If[DefinedQ[region, MagneticField], MagneticField@region, {0, 0, 0}],
        opts
    ];

(Hamiltonian[sys_?AtomicSystemQ, fields : {{genericOptionsPattern[]}..}, opts : OptionsPattern[]] := Module[
    {interactions = ProcessOption[Hamiltonian, Interaction, OptionValue[Hamiltonian, {#, opts}, Interaction]]& /@ fields,
        interaction},
    interactions = DeleteCases[interactions, "Internal", Infinity];
    interaction = Cases[interaction, "Internal"];
    Total[MapThread[Hamiltonian[sys, Interaction -> #1, #2, opts]&, {interactions, fields}]] +
        Hamiltonian[sys, Interaction -> interaction, opts] /;
        FreeQ[interactions, $Failed]
]) // ProcessOptions;

OptionProcessor[ElectricField] = OptionProcessor[MagneticField] =
    Switch[#,
        {{_, genericOptionsPattern[]}..},
        #,
        {_, genericOptionsPattern[]},
        {#},
        _,
        {{#}}
    ]&;

processInteractionOption[field_, opts___] :=
    ProcessOption[Hamiltonian, Interaction, OptionValue[Hamiltonian, FilterRules[{Rest@field, opts}, Interaction], Interaction]];

(Hamiltonian[sys_?AtomicSystemQ, opts : OptionsPattern[]] := Module[
    {interaction, electricField, magneticField},
    WignerEckart[sys, {Zero, 0, 0}]
        + If[MemberQ[interaction, "Internal"], InternalHamiltonian[sys, FilterRules@opts], 0]
        + Total@MapThread[
            Sum[If[MemberQ[processInteractionOption[field, opts], #1], #2[sys, First@field, FilterRules@{Rest@field, opts}], 0], {field, #3}]&,
            {
                {"ElectricDipole", "MagneticDipole", "Polarizability"},
                {E1Hamiltonian, MagneticHamiltonian, PolarizabilityHamiltonian},
                {electricField, magneticField, electricField}
            }
        ]
]) // ProcessOptions;

End[];

EndPackage[];
